from typing import Union
from os.path import join
from datetime import date
import numpy as np
import logging

import colored_logging as cl
from rasters import Raster

logger = logging.getLogger(__name__)

def diagnostic(values: Union[Raster, np.ndarray], variable: str, show_distributions: bool = True, output_directory: str = None):
    if isinstance(values, Raster) and output_directory is not None:
        filename = join(output_directory, f"{variable}.tif")
        logger.info(filename)
        values.to_geotiff(filename)

    if show_distributions:
        unique = np.unique(values)
        nan_proportion = np.count_nonzero(np.isnan(values)) / np.size(values)

        if len(unique) < 10:
            logger.info(f"variable {cl.name(variable)} ({values.dtype}) has {cl.val(unique)} unique values")

            for value in unique:
                if np.isnan(value):
                    count = np.count_nonzero(np.isnan(values))
                else:
                    count = np.count_nonzero(values == value)

                if value == 0 or np.isnan(value):
                    logger.info(f"* {cl.colored(value, 'red')}: {cl.colored(count, 'red')}")
                else:
                    logger.info(f"* {cl.val(value)}: {cl.val(count)}")
        else:
            minimum = np.nanmin(values)

            if minimum < 0:
                minimum_string = cl.colored(f"{minimum:0.3f}", "red")
            else:
                minimum_string = cl.val(f"{minimum:0.3f}")

            maximum = np.nanmax(values)

            if maximum <= 0:
                maximum_string = cl.colored(f"{maximum:0.3f}", "red")
            else:
                maximum_string = cl.val(f"{maximum:0.3f}")

            if nan_proportion > 0.5:
                nan_proportion_string = cl.colored(f"{(nan_proportion * 100):0.2f}%", "yellow")
            elif nan_proportion == 1:
                nan_proportion_string = cl.colored(f"{(nan_proportion * 100):0.2f}%", "red")
            else:
                nan_proportion_string = cl.val(f"{(nan_proportion * 100):0.2f}%")

            message = "variable " + cl.name(variable) + \
                " min: " + minimum_string + \
                " mean: " + cl.val(f"{np.nanmean(values):0.3f}") + \
                " max: " + maximum_string + \
                " nan: " + nan_proportion_string

            if np.all(values == 0):
                message += " all zeros"
                logger.warning(message)
            else:
                logger.info(message)

        if nan_proportion == 1:
            raise ValueError(f"variable {variable} is blank")



from typing import Union, Tuple
import numpy as np

import rasters as rt

from rasters import Raster

from .soil_heat_flux import calculate_SEBAL_soil_heat_flux

from .constants import *
from .canopy_air_stream import calculate_canopy_air_stream_vapor_pressure
from .soil_moisture_iteration import iterate_soil_moisture


def iterate_with_solar(
        seconds_of_day: Union[Raster, np.ndarray],  # Seconds of the day
        ST_C: Union[Raster, np.ndarray],  # Soil temperature (°C)
        NDVI: Union[Raster, np.ndarray],  # Normalized Difference Vegetation Index
        albedo: Union[Raster, np.ndarray],  # Albedo
        gB_ms: Union[Raster, np.ndarray],  # boundary layer conductance (m/s)
        gS_ms: Union[Raster, np.ndarray],  # stomatal conductance (m/s)
        LE_Wm2: Union[Raster, np.ndarray],  # latent heat flux (W/m^2)
        Rg_Wm2: Union[Raster, np.ndarray],  # Incoming solar radiation (W/m^2)
        Rn_Wm2: Union[Raster, np.ndarray],  # Net radiation (W/m^2)
        LWnet_Wm2: Union[Raster, np.ndarray],  # Net longwave radiation (W/m^2)
        Ta_C: Union[Raster, np.ndarray],  # Air temperature (°C)
        dTS_C: Union[Raster, np.ndarray],  # Change in soil temperature (°C)
        Td_C: Union[Raster, np.ndarray],  # Dew point temperature (°C)
        Tsd_C: Union[Raster, np.ndarray],  # Soil dew point temperature (°C)
        Ea_hPa: Union[Raster, np.ndarray],  # actual vapor pressure (hPa)
        Estar_hPa: Union[Raster, np.ndarray],  # saturation vapor pressure at surface temperature (hPa)
        VPD_hPa: Union[Raster, np.ndarray],  # Vapor pressure deficit (hPa)
        SVP_hPa: Union[Raster, np.ndarray],  # Saturation vapor pressure (hPa)
        delta_hPa: Union[Raster, np.ndarray],  # Slope of the saturation vapor pressure-temperature curve (hPa/°C)
        phi_Wm2: Union[Raster, np.ndarray],  # Net radiation minus soil heat flux (W/m^2)
        Es_hPa: Union[Raster, np.ndarray],  # Saturation vapor pressure (hPa)
        s1: Union[Raster, np.ndarray],  # Soil moisture parameter
        s3: Union[Raster, np.ndarray],  # Soil moisture parameter
        FVC: Union[Raster, np.ndarray],  # Fractional canopy cover
        T0_C: Union[Raster, np.ndarray],  # Reference temperature (°C)
        gB_by_gS: Union[Raster, np.ndarray],  # Ratio of boundary layer conductance to stomatal conductance
        rho_kgm3: Union[Raster, np.ndarray, float] = RHO_KGM3,  # Air density (kg/m^3)
        Cp_Jkg: Union[Raster, np.ndarray, float] = CP_JKG,  # Specific heat at constant pressure (J/kg/K)
        gamma_hPa: Union[Raster, np.ndarray, float] = GAMMA_HPA,  # Psychrometric constant (hPa/°C)
        G_method: str = "santanello"  # Method for calculating soil heat flux
    ) -> Tuple[Union[Raster, np.ndarray]]:
    """
    This function calculates the canopy air stream vapor pressures, vapor pressure deficit at source,
    vapor pressure at the reference height, soil moisture, soil heat flux, and recompute phi.

    Parameters:
    seconds_of_day (np.ndarray): Seconds of the day
    ST_C (np.ndarray): Soil temperature (°C)
    NDVI (np.ndarray): Normalized Difference Vegetation Index
    albedo (np.ndarray): Albedo
    gB (np.ndarray): Boundary layer conductance (m/s)
    gS (np.ndarray): Stomatal conductance (m/s)
    LE (np.ndarray): Latent heat flux (W/m^2)
    Rg (np.ndarray): Incoming solar radiation (W/m^2)
    Rn (np.ndarray): Net radiation (W/m^2)
    LWnet (np.ndarray): Net longwave radiation (W/m^2)
    Ta_C (np.ndarray): Air temperature (°C)
    dTS (np.ndarray): Change in soil temperature (°C)
    Td_C (np.ndarray): Dew point temperature (°C)
    Tsd_C (np.ndarray): Soil dew point temperature (°C)
    Ea_hPa (np.ndarray): Actual vapor pressure (hPa)
    Estar (np.ndarray): Saturation vapor pressure at surface temperature (hPa)
    VPD_hPa (np.ndarray): Vapor pressure deficit (hPa)
    SVP_hPa (np.ndarray): Saturation vapor pressure (hPa)
    delta (np.ndarray): Slope of the saturation vapor pressure-temperature curve (hPa/°C)
    phi (np.ndarray): Net radiation minus soil heat flux (W/m^2)
    Es (np.ndarray): Saturation vapor pressure (hPa)
    s1 (np.ndarray): Soil moisture parameter
    s3 (np.ndarray): Soil moisture parameter
    fc (np.ndarray): Fractional canopy cover
    T0 (np.ndarray): Reference temperature (°C)
    gB_by_gS (np.ndarray): Ratio of boundary layer conductance to stomatal conductance
    gamma (np.ndarray): Psychrometric constant (hPa/°C)
    rho (np.ndarray): Air density (kg/m^3)
    cp (np.ndarray): Specific heat at constant pressure (J/kg/K)
    G_method (str): Method for calculating soil heat flux (santanello or SEBAL)

    Returns:
    e0star (np.ndarray): Canopy air stream vapor pressures
    D0 (np.ndarray): Vapor pressure deficit at source
    e0 (np.ndarray): Vapor pressure at the reference height
    SM (np.ndarray): Soil moisture
    G (np.ndarray): Soil heat flux
    """
    # canopy air stream vapor pressures
    e0star = calculate_canopy_air_stream_vapor_pressure(
        LE=LE_Wm2,
        Ea_hPa=Ea_hPa,
        Estar=Estar_hPa,
        gB=gB_ms,
        gS=gS_ms
    )

    # vapor pressure deficit at source
    D0 = VPD_hPa + (delta_hPa * phi_Wm2 - (delta_hPa + gamma_hPa) * LE_Wm2) / (rho_kgm3 * Cp_Jkg * gB_ms)

    # Vapor pressure at the reference height (hPa)
    e0 = e0star - D0  
    e0 = rt.where(e0 < 0, Es_hPa, e0)
    e0 = rt.where(e0 > e0star, e0star, e0)

    # calculate soil moisture
    SM = iterate_soil_moisture(
        delta_hPa=delta_hPa,
        s1=s1,
        s3=s3,
        ST_C=ST_C,
        Ta_C=Ta_C,
        dTS_C=dTS_C,
        Td_C=Td_C,
        Tsd_C=Tsd_C,
        Rg_Wm2=Rg_Wm2,
        Rn_Wm2=Rn_Wm2,
        LWnet_Wm2=LWnet_Wm2,
        FVC=FVC,
        VPD_hPa=VPD_hPa,
        D0_hPa=D0,
        SVP_hPa=SVP_hPa,
        Ea_hPa=Ea_hPa,
        T0=T0_C,
        gamma_hPa=gamma_hPa
    )

    # calculate soil heat flux
    G = calculate_SEBAL_soil_heat_flux(
        ST_C=ST_C,
        NDVI=NDVI,
        albedo=albedo,
        Rn=Rn_Wm2
    )

    # recompute phi
    phi_Wm2 = Rn_Wm2 - G

    #
    alphaN = ((gS_ms * (e0star - Ea_hPa) * (2 * delta_hPa + 2 * gamma_hPa + gamma_hPa * gB_by_gS * (1 + SM))) / (
                2 * delta_hPa * (gamma_hPa * (T0_C - Ta_C) * (gB_ms + gS_ms) + gS_ms * (e0star - Ea_hPa))))

    return SM, G, e0, e0star, D0, alphaN



from typing import Union, Tuple
import numpy as np

import rasters as rt

from rasters import Raster

from .constants import *

def STIC_closure(
        delta_hPa: Union[Raster, np.ndarray],  # Slope of the saturation vapor pressure-temperature curve (hPa/°C)
        phi_Wm2: Union[Raster, np.ndarray],  # available energy (W/m^2)
        Es_hPa: Union[Raster, np.ndarray],  # Vapor pressure at the reference height (hPa)
        Ea_hPa: Union[Raster, np.ndarray],  # Actual vapor pressure (hPa)
        Estar_hPa: Union[Raster, np.ndarray],  # Saturation vapor pressure at the reference height (hPa)
        SM: Union[Raster, np.ndarray],  # Soil moisture (m³/m³)
        gamma_hPa: float = GAMMA_HPA,  # Psychrometric constant (hPa/°C)
        rho_kgm3: float = RHO_KGM3,  # Air density (kg/m³)
        Cp_Jkg: float = CP_JKG,  # Specific heat capacity of air (J/kg/°C)
        alpha: float = PT_ALPHA  # Priestley-Taylor alpha
    ) -> Tuple[Union[Raster, np.ndarray]]:
    """
    STIC closure equations with modified Priestley Taylor and Penman Monteith
    (Mallick et al., 2015, Water Resources research)

    Parameters:
    delta (np.ndarray): Slope of the saturation vapor pressure-temperature curve (hPa/°C)
    phi (np.ndarray): available energy (W/m^2)
    Es (np.ndarray): Vapor pressure at the reference height (hPa)
    Ea (np.ndarray): Actual vapor pressure (hPa)
    Estar (np.ndarray): Saturation vapor pressure at the reference height (hPa)
    SM (np.ndarray): Soil moisture (m³/m³)
    rho (float, optional): Air density (kg/m³). Defaults to RHO.
    cp (float, optional): Specific heat capacity of air (J/kg/°C). Defaults to CP.
    gamma (float, optional): Psychrometric constant (hPa/°C). Defaults to GAMMA.
    alpha (float, optional): Stability correction factor for conductance. Defaults to ALPHA.

    Returns:
    gB (np.ndarray): boundary layer conductance (m s^-1)
    gS (np.ndarray): stomatal conductance (m s^-1)
    dT (np.ndarray): difference between surface and air temperature
    EF (np.ndarray): evaporative fraction
    """
    # boundary layer conductance (m s^-1)
    gB = ((2 * phi_Wm2 * alpha * delta_hPa * gamma_hPa) / (2 * Cp_Jkg * delta_hPa * Es_hPa * rho_kgm3 - 2 * Cp_Jkg * delta_hPa * Ea_hPa * rho_kgm3 - 2 * Cp_Jkg * Ea_hPa * gamma_hPa * rho_kgm3 + Cp_Jkg * Es_hPa * gamma_hPa * rho_kgm3 + Cp_Jkg * Estar_hPa * gamma_hPa * rho_kgm3 - Cp_Jkg * SM * Es_hPa * gamma_hPa * rho_kgm3 + Cp_Jkg * SM * Estar_hPa * gamma_hPa * rho_kgm3))
    gB = rt.clip(gB, 0.0001, 0.2)

    # stomatal conductance (m s^-1)
    gS = (-(2 * (phi_Wm2 * alpha * delta_hPa * Ea_hPa * gamma_hPa - phi_Wm2 * alpha * delta_hPa * Es_hPa * gamma_hPa)) / (Cp_Jkg * Estar_hPa ** 2 * gamma_hPa * rho_kgm3 - Cp_Jkg * Es_hPa ** 2 * gamma_hPa * rho_kgm3 - 2 * Cp_Jkg * delta_hPa * Es_hPa ** 2 * rho_kgm3 + 2 * Cp_Jkg * delta_hPa * Ea_hPa * Es_hPa * rho_kgm3 - 2 * Cp_Jkg * delta_hPa * Ea_hPa * Estar_hPa * rho_kgm3 + 2 * Cp_Jkg * delta_hPa * Es_hPa * Estar_hPa * rho_kgm3 + 2 * Cp_Jkg * Ea_hPa * Es_hPa * gamma_hPa * rho_kgm3 - 2 * Cp_Jkg * Ea_hPa * Estar_hPa * gamma_hPa * rho_kgm3 + Cp_Jkg * SM * Es_hPa ** 2 * gamma_hPa * rho_kgm3 + Cp_Jkg * SM * Estar_hPa ** 2 * gamma_hPa * rho_kgm3 - 2 * Cp_Jkg * SM * Es_hPa * Estar_hPa * gamma_hPa * rho_kgm3))
    gS = rt.clip(gS, 0.0001, 0.2)

    # difference between surface and air temperature (Celsius)
    dT = (2 * delta_hPa * Es_hPa - 2 * delta_hPa * Ea_hPa - 2 * Ea_hPa * gamma_hPa + Es_hPa * gamma_hPa + Estar_hPa * gamma_hPa - SM * Es_hPa * gamma_hPa + SM * Estar_hPa * gamma_hPa + 2 * alpha * delta_hPa * Ea_hPa - 2 * alpha * delta_hPa * Es_hPa) / (2 * alpha * delta_hPa * gamma_hPa)
    dT = rt.clip(dT, -10, 50)

    # evaporative fraction
    EF = -(2 * alpha * delta_hPa * Ea_hPa - 2 * alpha * delta_hPa * Es_hPa) / (2 * delta_hPa * Es_hPa - 2 * delta_hPa * Ea_hPa - 2 * Ea_hPa * gamma_hPa + Es_hPa * gamma_hPa + Estar_hPa * gamma_hPa - SM * Es_hPa * gamma_hPa + SM * Estar_hPa * gamma_hPa)
    EF = rt.clip(EF, 0, 1)

    return gB, gS, dT, EF



from typing import Union, Tuple
import numpy as np

import rasters as rt
from rasters import Raster

from .vegetation_conversion.vegetation_conversion import FVC_from_NDVI

from .constants import GAMMA_HPA
from .root_zone_initialization import calculate_root_zone_moisture

def initialize_soil_moisture(
        delta_hPa: Union[Raster, np.ndarray],  # Rate of change of saturation vapor pressure with temperature (hPa/°C)
        ST_C: Union[Raster, np.ndarray],  # Surface temperature (°C)
        Ta_C: Union[Raster, np.ndarray],  # Air temperature (°C)
        Td_C: Union[Raster, np.ndarray],  # Dewpoint temperature (°C)
        dTS: Union[Raster, np.ndarray],  # Difference between surface and air temperature (°C)
        Rn_Wm2: Union[Raster, np.ndarray],  # Net radiation (W/m²)
        LWnet_Wm2: Union[Raster, np.ndarray],  # Net longwave radiation (W/m²)
        FVC: Union[Raster, np.ndarray],  # Fractional vegetation cover (unitless)
        VPD_hPa: Union[Raster, np.ndarray],  # Vapor pressure deficit (hPa)
        SVP_hPa: Union[Raster, np.ndarray],  # Saturation vapor pressure (hPa)
        Ea_hPa: Union[Raster, np.ndarray],  # Actual vapor pressure (hPa)
        Estar_hPa: Union[Raster, np.ndarray],  # Saturation vapor pressure at surface temperature (hPa)
        gamma_hPa: Union[Raster, np.ndarray, float] = GAMMA_HPA  # Psychrometric constant (hPa/°C)
    ) -> Tuple[Union[Raster, np.ndarray]]:
    """
    This function estimates the soil moisture availability (SM) based on thermal IR and meteorological information. 
    The estimated SM is treated as initial SM, which will be later estimated through iteration in the actual ET estimation loop to establish feedback between SM and biophysical states.

    Parameters:
    delta (np.ndarray): Rate of change of saturation vapor pressure with temperature (kPa/°C)
    ST_C (np.ndarray): Surface temperature (°C)
    Ta_C (np.ndarray): Air temperature (°C)
    Td_C (np.ndarray): Dewpoint temperature (°C)
    dTS (np.ndarray): Difference between surface and air temperature (°C)
    Rn (np.ndarray): Net radiation (W/m²)
    LWnet (np.ndarray): Net longwave radiation (W/m²)
    NDVI (np.ndarray): Normalized Difference Vegetation Index (unitless)
    VPD_hPa (np.ndarray): Vapor pressure deficit (hPa)
    SVP_hPa (np.ndarray): Saturation vapor pressure (hPa)
    Ea_hPa (np.ndarray): Actual vapor pressure (hPa)
    Estar (np.ndarray): Saturation vapor pressure at surface temperature (hPa)

    Returns:
    SM (np.ndarray): soil moisture (m³/m³)
    Mrz (np.ndarray): The rootzone moisture (m³/m³)
    Ms (np.ndarray): The surface moisture (m³/m³)
    Ep_PT (np.ndarray): The potential evaporation (Priestley-Taylor eqn.) (mm/day)
    Ds (np.ndarray): The vapor pressure deficit at surface (hPa)
    s1 (np.ndarray): The slope of SVP at dewpoint temperature (hPa/K)
    s3 (np.ndarray): The slope of SVP at surface temperature (hPa/K)
    Tsd_C (np.ndarray): The surface dewpoint temperature (°C)
    """
    # Compute the surface dewpoint temperature
    s11 = (45.03 + 3.014 * Td_C + 0.05345 * Td_C ** 2 + 0.00224 * Td_C ** 3) * 1e-2  # slope of SVP at TD (hpa/K)
    s22 = (Estar_hPa - Ea_hPa) / (ST_C - Td_C)
    s33 = (45.03 + 3.014 * ST_C + 0.05345 * ST_C ** 2 + 0.00224 * ST_C ** 3) * 1e-2  # slope of SVP at TS (hpa/K)
    s44 = (SVP_hPa - Ea_hPa) / (Ta_C - Td_C)
    Tsd_C = (Estar_hPa - Ea_hPa - s33 * ST_C + s11 * Td_C) / (s11 - s33)  # Surface dewpoint temperature (degC)

    # Calculate the surface moisture or wetness
    Msurf = (s11 / s22) * ((Tsd_C - Td_C) / (ST_C - Td_C))  # Surface wetness
    Msurf = rt.clip(Msurf, 0.0001, 0.9999)

    # Calculate the surface vapor pressure and deficit
    esurf = Ea_hPa + Msurf * (Estar_hPa - Ea_hPa)
    Dsurf = esurf - Ea_hPa

    # Separate the soil and canopy wetness to form a composite surface moisture
    Ms = Msurf
    Mcan = FVC * Msurf
    Msoil = (1 - FVC) * Msurf

    TdewIndex = (ST_C - Tsd_C) / (Ta_C - Td_C)  # % TdewIndex > 1 signifies super dry condition
    Ep_PT = (1.26 * delta_hPa * Rn_Wm2) / (delta_hPa + gamma_hPa)  # Potential evaporation (Priestley-Taylor eqn.)

    # Adjust surface wetness based on certain conditions
    Ms = rt.where((FVC <= 0.25) & (TdewIndex < 1), Msoil, Ms)
    Mcan = rt.where((FVC <= 0.25) & (TdewIndex < 1), 0, Mcan)
    Ms = rt.where((FVC <= 0.25) & (Ta_C > 10) & (Td_C < 0) & (LWnet_Wm2 < -125), Msoil, Ms)
    Mcan = rt.where((FVC <= 0.25) & (Ta_C > 10) & (Td_C < 0) & (LWnet_Wm2 < -125), 0, Mcan)
    
    # Calculate the root-zone moisture
    SMrz = calculate_root_zone_moisture(
        delta_hPa=delta_hPa, 
        ST_C=ST_C, 
        Ta_C=Ta_C, 
        Td_C=Td_C, 
        s11=s11, 
        s33=s33, 
        s44=s44, 
        Tsd_C=Tsd_C,
        gamma_hPa=gamma_hPa
    )

    # Combine soil moisture to account for hysteresis and initial estimation of surface vapor pressure
    SM = Ms
    SM = rt.where((Ep_PT > Rn_Wm2) & (dTS > 0), SMrz, SM)
    SM = rt.where((Ep_PT > Rn_Wm2) & (FVC <= 0.25), SMrz, SM)
    SM = rt.where((Ep_PT > Rn_Wm2) & (Dsurf > VPD_hPa), SMrz, SM)
    SM = rt.where((FVC <= 0.25) & (dTS > 0) & (Ta_C > 10) & (Td_C < 0) & (LWnet_Wm2 < -125), SMrz, SM)
    SM = rt.where((FVC <= 0.25) & (dTS > 0) & (Ta_C > 10) & (Td_C < 0) & (Dsurf > VPD_hPa), SMrz, SM)
    SM = rt.where((Ep_PT < Rn_Wm2) & (FVC <= 0.25) & (Dsurf > VPD_hPa), SMrz, SM)
    
    es = Ea_hPa + SM * (Estar_hPa - Ea_hPa)
    
    # vapor pressure deficit at surface
    Ds = (Estar_hPa - es)  
    
    s1 = s11
    s3 = s33

    return SM, SMrz, Ms, Ep_PT, Ds, s1, s3, Tsd_C



from typing import Union, Tuple
import numpy as np

import rasters as rt
from rasters import Raster

from .constants import *
from .canopy_air_stream import calculate_canopy_air_stream_vapor_pressure
from .root_zone_initialization import calculate_root_zone_moisture

def iterate_without_solar(
        LE: Union[Raster, np.ndarray],  # Latent heat flux (W/m^2)
        PET: Union[Raster, np.ndarray],  # Potential evapotranspiration (W/m^2)
        SM: Union[Raster, np.ndarray],
        ST_C: Union[Raster, np.ndarray],  # Surface temperature (°C)
        Ta_C: Union[Raster, np.ndarray],  # Air temperature (°C)
        dTS: Union[Raster, np.ndarray],  # Surface-air temperature difference (°C)
        T0: Union[Raster, np.ndarray],  # Reference temperature (°C)
        gB: Union[Raster, np.ndarray],  # Boundary layer conductance (m/s)
        gS: Union[Raster, np.ndarray],  # Stomatal conductance (m/s)
        Ea_hPa: Union[Raster, np.ndarray],  # Actual vapor pressure (hPa)
        Td_C: Union[Raster, np.ndarray],  # Dew point temperature (°C)
        VPD_hPa: Union[Raster, np.ndarray],  # Vapor pressure deficit (hPa)
        Estar: Union[Raster, np.ndarray],  # Saturation vapor pressure at surface temperature (hPa)
        delta: Union[Raster, np.ndarray],  # Slope of the saturation vapor pressure-temperature curve (hPa/°C)
        phi: Union[Raster, np.ndarray],  # available energy (W/m^2)
        Ds: Union[Raster, np.ndarray],  # Vapor pressure deficit at source (hPa)
        Es: Union[Raster, np.ndarray],  # Saturation vapor pressure (hPa)
        s3: Union[Raster, np.ndarray],  # Slope of the saturation vapor pressure and temperature 
        s4: Union[Raster, np.ndarray],  # Slope of the saturation vapor pressure and temperature 
        gB_by_gS: Union[Raster, np.ndarray],  # Ratio of boundary layer conductance to stomatal conductance
        gamma_hPa: Union[Raster, np.ndarray] = GAMMA_HPA,  # Psychrometric constant (hPa/°C)
        rho_kgm3: Union[Raster, np.ndarray] = RHO_KGM3,  # Air density (kg/m^3)
        Cp_Jkg: Union[Raster, np.ndarray] = CP_JKG,  # Specific heat at constant pressure (J/kg/K)
    ) -> Tuple[Union[Raster, np.ndarray]]:
    """
    Iterate STIC model without knowing solar radiation.

    Parameters:
    LE (np.ndarray): Latent heat flux (W/m^2)
    PET (np.ndarray): Potential evapotranspiration (W/m^2)
    ST_C (np.ndarray): Surface temperature (°C)
    Ta_C (np.ndarray): Air temperature (°C)
    dTS (np.ndarray): Surface-air temperature difference (°C)
    T0 (np.ndarray): Reference temperature (°C)
    gB (np.ndarray): Boundary layer conductance (m/s)
    gS (np.ndarray): Stomatal conductance (m/s)
    Ea_hPa (np.ndarray): Actual vapor pressure (hPa)
    Td_C (np.ndarray): Dew point temperature (°C)
    VPD_hPa (np.ndarray): Vapor pressure deficit (hPa)
    Estar (np.ndarray): Saturation vapor pressure at surface temperature (hPa)
    delta (np.ndarray): Slope of the saturation vapor pressure-temperature curve (hPa/°C)
    phi (np.ndarray): available energy (W/m^2)
    Ds (np.ndarray): Vapor pressure deficit at source (hPa)
    Es (np.ndarray): Saturation vapor pressure (hPa)
    s3 (np.ndarray): Slope of the saturation vapor pressure and temperature 
    s4 (np.ndarray): Soil moisture parameter
    gB_by_gS (np.ndarray): Ratio of boundary layer conductance to stomatal conductance
    gamma (np.ndarray): Psychrometric constant (hPa/°C)
    rho (np.ndarray): Air density (kg/m^3)
    cp (np.ndarray): Specific heat at constant pressure (J/kg/K)
    
    Returns:
    SM (np.ndarray): soil moisture (m³/m³)
    SMrz (np.ndarray): Root zone moisture (m³/m³)
    Ms (np.ndarray): Surface Moisture
    s1 (np.ndarray): lope of the saturation vapor pressure and temperature
    e0 (np.ndarray): Vapor pressure at the reference height (hPa)
    e0star (np.ndarray): Canopy air stream vapor pressures (hPa)
    Tsd_C (np.ndarray): Soil dew point temperature (°C)
    D0 (np.ndarray): Vapor pressure deficit at source (hPa)
    alphaN (np.ndarray): Alpha parameter
    """

    # canopy air stream vapor pressures
    e0star = calculate_canopy_air_stream_vapor_pressure(
        LE=LE, # latent heat flux (W/m^2)
        Ea_hPa=Ea_hPa, # actual vapor pressure (hPa)
        Estar=Estar, # saturation vapor pressure at surface temperature (hPa)
        gB=gB, # boundary layer conductance (m/s)
        gS=gS # stomatal conductance (m/s),
        
    )

    # vapor pressure deficit at source
    D0 = VPD_hPa + (delta * phi - (delta + gamma_hPa) * LE) / (rho_kgm3 * Cp_Jkg * gB)  
    D0 = rt.where(D0 < 0, Ds, D0)

    # Vapor pressure at the reference height (hPa)
    e0 = rt.clip(e0star - D0, Es, e0star)

    # re-estimating M (direct LST feedback into M computation)
    s1 = (45.03 + 3.014 * Td_C + 0.05345 * Td_C ** 2 + 0.00224 * Td_C ** 3) * 1e-2  # Soil moisture parameter
    Tsd_C = Td_C + (gamma_hPa * LE) / (rho_kgm3 * Cp_Jkg * gB * s1)  # Soil dew point temperature (°C)

    # Surface Moisture Ms
    Ms = rt.clip(s1 * (Tsd_C - Td_C) / (s3 * (ST_C - Td_C)), 0, 1)

    # Root zone moisture Mrz
    SMrz = calculate_root_zone_moisture(
        delta_hPa=delta,  # Slope of the saturation vapor pressure-temperature curve (hPa/°C)
        ST_C=ST_C,  # Surface temperature (°C)
        Ta_C=Ta_C,  # Air temperature (°C)
        Td_C=Td_C,  # Dew point temperature (°C)
        s11=s1,  # Soil moisture parameter
        s33=s3,  # Soil moisture parameter
        s44=s4,  # Soil moisture parameter
        Tsd_C=Tsd_C  # Soil dew point temperature (°C)
    )

    # combining hysteresis logic to differentiate surface vs. rootzone water control
    SM = rt.where((D0 > VPD_hPa) & (PET > phi) & (dTS > 0), SMrz, SM)
    SM = rt.where((phi > 0) & (dTS > 0) & (Td_C <= 0), SMrz, SM)
    SM = rt.clip(SM, 0, 1)

    # checking convergence
    # re-estimating alpha
    alphaN = ((gS * (e0star - Ea_hPa) * (2 * delta + 2 * gamma_hPa + gamma_hPa * gB_by_gS * (1 + SM))) / (
            2 * delta * (gamma_hPa * (T0 - Ta_C) * (gB + gS) + gS * (e0star - Ea_hPa))))

    return SM, SMrz, Ms, s1, e0, e0star, Tsd_C, D0, alphaN



from typing import Union
import numpy as np

import rasters as rt

from rasters import Raster

from .constants import GAMMA_HPA

def calculate_rootzone_moisture(
        delta_hPa: Union[Raster, np.ndarray], # Rate of change of saturation vapor pressure with temperature (hPa/°C)
        s1_hPa: Union[Raster, np.ndarray], # The slope of SVP at surface temperature (hPa/K)
        s3_hPa: Union[Raster, np.ndarray], # The slope of SVP at dewpoint temperature (hPa/K)
        ST_C: Union[Raster, np.ndarray], # surface temperature (°C)
        Ta_C: Union[Raster, np.ndarray], # air temperature (°C)
        dTS_C: Union[Raster, np.ndarray], # difference between surface and air temperature (°C)
        Td_C: Union[Raster, np.ndarray], # dewpoint temperature (°C)
        Tsd_C: Union[Raster, np.ndarray], # surface dewpoint temperature (°C)
        Rg_Wm2: Union[Raster, np.ndarray], # Incoming solar radiation (W/m^2)
        Rn_Wm2: Union[Raster, np.ndarray], # Net radiation (W/m^2)
        LWnet_Wm2: Union[Raster, np.ndarray], # Net longwave radiation (W/m^2)
        FVC: Union[Raster, np.ndarray], # Fractional vegetation cover (unitless)
        VPD_hPa: Union[Raster, np.ndarray], # Vapor pressure deficit (hPa)
        D0_hPa: Union[Raster, np.ndarray], # Vapor pressure deficit at source (hPa)
        SVP_hPa: Union[Raster, np.ndarray], # Saturation vapor pressure (hPa)
        Ea_hPa: Union[Raster, np.ndarray], # Actual vapor pressure (hPa)
        gamma_hPa: Union[Raster, np.ndarray, float] = GAMMA_HPA # Psychrometric constant (hPa/°C)
    ) -> Union[Raster, np.ndarray]:
    """
    Calculates the root zone moisture (Mrz) based on thermal IR and meteorological information.

    Args:
        delta (np.ndarray): Rate of change of saturation vapor pressure with temperature (kPa/°C).
        s1 (np.ndarray): The slope of saturation vapor pressure at surface temperature (hPa/K).
        s3 (np.ndarray): The slope of saturation vapor pressure at dewpoint temperature (hPa/K).
        ST_C (np.ndarray): Surface temperature in degrees Celsius.
        Ta_C (np.ndarray): Air temperature in degrees Celsius.
        dTS (np.ndarray): Difference between surface and air temperature in degrees Celsius.
        Td_C (np.ndarray): Dewpoint temperature in degrees Celsius.
        Tsd_C (np.ndarray): Surface dewpoint temperature in degrees Celsius.
        Rg (np.ndarray): Incoming solar radiation in W/m^2.
        Rn (np.ndarray): Net radiation in W/m^2.
        Lnet (np.ndarray): Net longwave radiation in W/m^2.
        fc (np.ndarray): Fractional vegetation cover (unitless).
        VPD_hPa (np.ndarray): Vapor pressure deficit in hPa.
        D0 (np.ndarray): Vapor pressure deficit at source in hPa.
        SVP_hPa (np.ndarray): Saturation vapor pressure in hPa.
        Ea_hPa (np.ndarray): Actual vapor pressure in hPa.

    Returns:
        np.ndarray: Root zone moisture (Mrz) (value 0 to 1).
    """
    s44 = (SVP_hPa - Ea_hPa) / (Ta_C - Td_C)
    kTSTD = (ST_C - Tsd_C) / (Ta_C - Td_C)
    Mrz = (gamma_hPa * s1_hPa * (Tsd_C - Td_C)) / (delta_hPa * s3_hPa * kTSTD * (ST_C - Td_C) + gamma_hPa * s44 * (Ta_C - Td_C) - delta_hPa * s1_hPa * (Tsd_C - Td_C))
    Mrz = rt.where((Rn_Wm2 < 0) & (dTS_C < 0) & (Mrz < 0), np.abs(Mrz), Mrz)
    Mrz = rt.where((Rn_Wm2 < 0) & (dTS_C > 0) & (Mrz < 0), np.abs(Mrz), Mrz)
    Mrz = rt.where((Rn_Wm2 > 0) & (dTS_C < 0) & (Mrz < 0), np.abs(Mrz), Mrz)
    Mrz = rt.where((Rn_Wm2 > 0) & (dTS_C > 0) & (Mrz < 0), np.abs(Mrz), Mrz)
    Mrz = rt.where((Rg_Wm2 > 0) & (Mrz < 0), np.abs(Mrz), Mrz)
    Mrz = rt.where((Rg_Wm2 < 0) & (Mrz < 0), np.abs(Mrz), Mrz)
    Mrz = rt.where((Td_C < 0) & (Mrz < 0), np.abs(Mrz), Mrz)
    Mrz = rt.where(Mrz > 1, 1, Mrz)
    Mrz = rt.where(Mrz < 0, 0.0001, Mrz)
    
    return Mrz


from typing import Union
import numpy as np

import rasters as rt

from rasters import Raster

from .constants import GAMMA_HPA
from .root_zone_iteration import calculate_rootzone_moisture

def iterate_soil_moisture(
        delta_hPa: Union[Raster, np.ndarray], # Rate of change of saturation vapor pressure with temperature (hPa/°C)
        s1: Union[Raster, np.ndarray], # The slope of SVP at surface temperature (hPa/K)
        s3: Union[Raster, np.ndarray], # The slope of SVP at dewpoint temperature (hPa/K)
        ST_C: Union[Raster, np.ndarray], # surface temperature (°C)
        Ta_C: Union[Raster, np.ndarray], # air temperature (°C)
        dTS_C: Union[Raster, np.ndarray], # difference between surface and air temperature (°C)
        Td_C: Union[Raster, np.ndarray], # dewpoint temperature (°C)
        Tsd_C: Union[Raster, np.ndarray], # surface dewpoint temperature (°C)
        Rg_Wm2: Union[Raster, np.ndarray], # Incoming solar radiation (W/m^2)
        Rn_Wm2: Union[Raster, np.ndarray], # Net radiation (W/m^2)
        LWnet_Wm2: Union[Raster, np.ndarray], # Net longwave radiation (W/m^2)
        FVC: Union[Raster, np.ndarray], # Fractional vegetation cover (unitless)
        VPD_hPa: Union[Raster, np.ndarray], # Vapor pressure deficit (hPa)
        D0_hPa: Union[Raster, np.ndarray], # Vapor pressure deficit at source (hPa)
        SVP_hPa: Union[Raster, np.ndarray], # Saturation vapor pressure (hPa)
        Ea_hPa: Union[Raster, np.ndarray], # Actual vapor pressure (hPa)
        T0: Union[Raster, np.ndarray], # Temperature at source (°C)
        gamma_hPa: Union[Raster, np.ndarray, float] = GAMMA_HPA # Psychrometric constant (hPa/°C)
    ) -> Union[Raster, np.ndarray]:
    """
    Estimates the soil moisture availability (SM) (or wetness) (value 0 to 1) based on thermal IR and meteorological
    information. However, this M will be treated as initial M, which will be later on estimated through iteration in the
    actual ET estimation loop to establish feedback between M and biophysical states.

    Args:
        delta (np.ndarray): Rate of change of saturation vapor pressure with temperature (kPa/°C).
        s1 (np.ndarray): The slope of saturation vapor pressure at surface temperature (hPa/K).
        s3 (np.ndarray): The slope of saturation vapor pressure at dewpoint temperature (hPa/K).
        ST_C (np.ndarray): Surface temperature in degrees Celsius.
        Ta_C (np.ndarray): Air temperature in degrees Celsius.
        dTS (np.ndarray): Difference between surface and air temperature in degrees Celsius.
        Td_C (np.ndarray): Dewpoint temperature in degrees Celsius.
        Tsd_C (np.ndarray): Surface dewpoint temperature in degrees Celsius.
        Rg (np.ndarray): Incoming solar radiation in W/m^2.
        Rn (np.ndarray): Net radiation in W/m^2.
        LWnet (np.ndarray): Net longwave radiation in W/m^2.
        FVC (np.ndarray): Fractional vegetation cover (unitless).
        VPD_hPa (np.ndarray): Vapor pressure deficit in hPa.
        D0 (np.ndarray): Vapor pressure deficit at source in hPa.
        SVP_hPa (np.ndarray): Saturation vapor pressure in hPa.
        Ea_hPa (np.ndarray): Actual vapor pressure in hPa.
        T0 (np.ndarray): Temperature at source in degrees Celsius.

    Returns:
        np.ndarray: Soil moisture availability (SM) (or wetness) (value 0 to 1).
    """
    # calculate surface wetness (Msurf)
    kTSTD = (T0 - Td_C) / (ST_C - Td_C)
    Msurf = (s1 / s3) * ((Tsd_C - Td_C) / (kTSTD * (ST_C - Td_C)))  # Surface wetness
    Msurf = rt.where((Rn_Wm2 < 0) & (dTS_C < 0) & (Msurf < 0), np.abs(Msurf), Msurf)
    Msurf = rt.where((Rn_Wm2 < 0) & (dTS_C > 0) & (Msurf < 0), np.abs(Msurf), Msurf)
    Msurf = rt.where((Rn_Wm2 > 0) & (dTS_C < 0) & (Msurf < 0), np.abs(Msurf), Msurf)
    Msurf = rt.where((Rn_Wm2 > 0) & (dTS_C > 0) & (Msurf < 0), np.abs(Msurf), Msurf)
    
    # solar radiation is only used to correct negative values of surface wetness
    Msurf = rt.where((Rg_Wm2 > 0) & (Msurf < 0), np.abs(Msurf), Msurf)
    Msurf = rt.where((Rg_Wm2 < 0) & (Msurf < 0), np.abs(Msurf), Msurf)
    
    Msurf = rt.where((Td_C < 0) & (Msurf < 0), np.abs(Msurf), Msurf)
    Msurf = rt.clip(Msurf, 0.0001, 1)

    # Separating soil and canopy wetness to form a composite surface moisture
    Ms = Msurf
    Mcan = FVC * Msurf
    Msoil = (1 - FVC) * Msurf

    TdewIndex = (ST_C - Tsd_C) / (Ta_C - Td_C)

    # Potential evaporation (Priestley-Taylor eqn.)
    Ep_PT = (1.26 * delta_hPa * Rn_Wm2) / (delta_hPa + gamma_hPa)  

    # calculate surface wetness (Ms)
    Ms = rt.where((Ep_PT > Rn_Wm2) & (FVC <= 0.25) & (TdewIndex < 1), Msoil, Ms)
    Ms = rt.where((FVC <= 0.25) & (Ta_C > 10) & (Td_C < 0) & (LWnet_Wm2 < -125), Msoil, Ms)
    Ms = rt.where((Rn_Wm2 > Ep_PT) & (FVC <= 0.25) & (TdewIndex < 1) & (Td_C <= 0), Msoil, Ms)
    Ms = rt.where((Rn_Wm2 > Ep_PT) & (FVC <= 0.25) & (TdewIndex < 1), Msoil, Ms)
    Ms = rt.where((D0_hPa > VPD_hPa) & (FVC <= 0.25) & (TdewIndex < 1), Msoil, Ms)

    # calculate canopy wetness (Mcan)
    Mcan = rt.where((Ep_PT > Rn_Wm2) & (FVC <= 0.25) & (TdewIndex < 1), 0, Mcan)
    Mcan = rt.where((FVC <= 0.25) & (Ta_C > 10) & (Td_C < 0) & (LWnet_Wm2 < -125), 0, Mcan)
    Mcan = rt.where((Rn_Wm2 > Ep_PT) & (FVC <= 0.25) & (TdewIndex < 1) & (Td_C <= 0), 0, Mcan)
    Mcan = rt.where((Rn_Wm2 > Ep_PT) & (FVC <= 0.25) & (TdewIndex < 1), 0, Mcan)
    Mcan = rt.where((D0_hPa > VPD_hPa) & (FVC <= 0.25) & (TdewIndex < 1), 0, Mcan)

    # calculate rootzone moisture (Mrz)
    Mrz = calculate_rootzone_moisture(
        delta_hPa=delta_hPa, 
        s1_hPa=s1, 
        s3_hPa=s3, 
        ST_C=ST_C, 
        Ta_C=Ta_C, 
        dTS_C=dTS_C, 
        Td_C=Td_C, 
        Tsd_C=Tsd_C, 
        Rg_Wm2=Rg_Wm2, 
        Rn_Wm2=Rn_Wm2, 
        LWnet_Wm2=LWnet_Wm2, 
        FVC=FVC, 
        VPD_hPa=VPD_hPa, 
        D0_hPa=D0_hPa, 
        SVP_hPa=SVP_hPa, 
        Ea_hPa=Ea_hPa,
        gamma_hPa=gamma_hPa
    )

    TdewIndex = (ST_C - Tsd_C) / (Ta_C - Td_C)

    # Potential evaporation (Priestley-Taylor eqn.)
    Ep_PT = (1.26 * delta_hPa * Rn_Wm2) / (delta_hPa + gamma_hPa)  

    # COMBINE M to account for Hysteresis and initial estimation of surface vapor pressure
    SM = Msurf
    SM = rt.where((Ep_PT > Rn_Wm2) & (dTS_C > 0) & (FVC <= 0.25) & (D0_hPa > VPD_hPa) & (TdewIndex < 1), Mrz, SM)
    SM = rt.where((FVC <= 0.25) & (dTS_C > 0) & (Ta_C > 10) & (Td_C < 0) & (LWnet_Wm2 < -125) & (D0_hPa > VPD_hPa), Mrz, SM)

    return SM






from typing import Union
import numpy as np

import rasters as rt
from rasters import Raster

from .constants import GAMMA_HPA

def calculate_root_zone_moisture(
        delta_hPa: Union[Raster, np.ndarray],
        ST_C: Union[Raster, np.ndarray], 
        Ta_C: Union[Raster, np.ndarray], 
        Td_C: Union[Raster, np.ndarray], 
        s11: Union[Raster, np.ndarray], 
        s33: Union[Raster, np.ndarray], 
        s44: Union[Raster, np.ndarray], 
        Tsd_C: Union[Raster, np.ndarray],
        gamma_hPa: Union[Raster, np.ndarray, float] = GAMMA_HPA) -> Union[Raster, np.ndarray]:
    """
    This function calculates the rootzone moisture (Mrz) using various parameters.

    Parameters:
    delta (np.ndarray): Rate of change of saturation vapor pressure with temperature (hPa/°C)
    ST_C (np.ndarray): Surface temperature (°C)
    Ta_C (np.ndarray): Air temperature (°C)
    Td_C (np.ndarray): Dewpoint temperature (°C)
    s11 (np.ndarray): The slope of SVP at dewpoint temperature (hPa/K)
    s33 (np.ndarray): The slope of SVP at surface temperature (hPa/K)
    s44 (np.ndarray): The difference between saturation vapor pressure and actual vapor pressure divided by the difference between air temperature and dewpoint temperature (hPa/K)
    Tsd_C (np.ndarray): The surface dewpoint temperature (°C)

    Returns:
    SMrz (np.ndarray): The rootzone moisture (m³/m³)
    """
    return rt.clip((gamma_hPa * s11 * (Tsd_C - Td_C)) / (delta_hPa * s33 * (ST_C - Td_C) + gamma_hPa * s44 * (Ta_C - Td_C) - delta_hPa * s11 * (Tsd_C - Td_C)), 0.0001, 0.999)
    


from typing import Union
import numpy as np

import rasters as rt

from rasters import Raster

from .constants import *

def calculate_canopy_air_stream_vapor_pressure(
        LE: Union[Raster, np.ndarray],  # Latent heat flux [W/m^2]
        Ea_hPa: Union[Raster, np.ndarray],  # Actual vapor pressure [hPa]
        Estar: Union[Raster, np.ndarray],  # Saturation vapor pressure [hPa]
        gB: Union[Raster, np.ndarray],  # Conductance of boundary layer [mol/m^2/s]
        gS: Union[Raster, np.ndarray],  # Conductance of stomata [mol/m^2/s]
        rho_kgm3: Union[Raster, np.ndarray, float] = RHO_KGM3,  # Air density (kg/m^3)
        Cp_Jkg: Union[Raster, np.ndarray, float] = CP_JKG,  # Specific heat at constant pressure (J/kg/K)
        gamma_hPa: Union[Raster, np.ndarray, float] = GAMMA_HPA,  # Psychrometric constant (hPa/°C)
    ) -> Union[Raster, np.ndarray]:
    """
    Calculate the canopy air stream vapor pressure.

    Parameters:
        LE (Union[Raster, np.ndarray]): Latent heat flux [W/m^2]
        Ea_hPa (Union[Raster, np.ndarray]): Actual vapor pressure [hPa]
        Estar (Union[Raster, np.ndarray]): Saturation vapor pressure [hPa]
        gB (Union[Raster, np.ndarray]): Conductance of boundary layer [mol/m^2/s]
        gS (Union[Raster, np.ndarray]): Conductance of stomata [mol/m^2/s]

    Returns:
        Union[Raster, np.ndarray]: The calculated canopy air stream vapor pressure [hPa]
    """
    e0star = Ea_hPa + (gamma_hPa * LE * (gB + gS)) / (rho_kgm3 * Cp_Jkg * gB * gS)
    e0star = rt.where(e0star < 0, Estar, e0star)
    e0star = rt.where(e0star > 250, Estar, e0star)

    return e0star



from typing import Union, Tuple
import numpy as np

import rasters as rt

from rasters import Raster

from .vegetation_conversion.vegetation_conversion import FVC_from_NDVI
from .soil_heat_flux import calculate_SEBAL_soil_heat_flux

from .constants import *
from .soil_moisture_initialization import initialize_soil_moisture
from .net_radiation import calculate_net_longwave_radiation


def initialize_with_solar(
        seconds_of_day: Union[Raster, np.ndarray],  # time of day in seconds since midnight
        Rg_Wm2: Union[Raster, np.ndarray],  # solar radiation (W/m^2)
        Rn_Wm2: Union[Raster, np.ndarray],  # net radiation (W/m^2)
        ST_C: Union[Raster, np.ndarray],  # surface temperature (Celsius)
        emissivity: Union[Raster, np.ndarray],  # emissivity of the surface
        Ta_C: Union[Raster, np.ndarray],  # air temperature (Celsius)
        dTS_C: Union[Raster, np.ndarray],  # surface air temperature difference (Celsius)
        Td_C: Union[Raster, np.ndarray],  # dew point temperature (Celsius)
        VPD_hPa: Union[Raster, np.ndarray],  # vapor pressure deficit (hPa)
        SVP_hPa: Union[Raster, np.ndarray],  # saturation vapor pressure at given air temperature (hPa)
        Ea_hPa: Union[Raster, np.ndarray],  # actual vapor pressure at air temperature (hPa)
        Estar_hPa: Union[Raster, np.ndarray],  # saturation vapor pressure at surface temperature (hPa)
        delta_hPa: Union[Raster, np.ndarray],  # slope of saturation vapor pressure to air temperature (hpa/K)
        NDVI: Union[Raster, np.ndarray],  # normalized difference vegetation index
        FVC: Union[Raster, np.ndarray],  # fractional vegetation cover
        LAI: Union[Raster, np.ndarray],  # leaf area index
        albedo: Union[Raster, np.ndarray],  # albedo of the surface
        gamma_hPa: Union[Raster, np.ndarray, float] = GAMMA_HPA,  # psychrometric constant (hPa/°C)
        G_method: str = DEFAULT_G_METHOD  # method for calculating soil heat flux
    ) -> Tuple[Union[Raster, np.ndarray]]:
    # Rn SOIL
    kRN = 0.6
    Rn_soil = Rn_Wm2 * np.exp(-kRN * LAI)

    LWnet = calculate_net_longwave_radiation(
        Ta_C=Ta_C, 
        Ea_hPa=Ea_hPa, 
        ST_C=ST_C, 
        emissivity=emissivity, 
        albedo=albedo
    )
    
    # initialize soil moisture
    SM, SMrz, Ms, Ep_PT, Ds, s1, s3, Tsd_C = initialize_soil_moisture(
        delta_hPa=delta_hPa,  # slope of saturation vapor pressure to air temperature (hpa/K)
        ST_C=ST_C,  # surface temperature (Celsius)
        Ta_C=Ta_C,  # air temperature (Celsius)
        Td_C=Td_C,  # dew point temperature (Celsius)
        dTS=dTS_C,  # surface air temperature difference (Celsius)
        Rn_Wm2=Rn_Wm2,  # net radiation (W/m^2)
        LWnet_Wm2=LWnet,  # net longwave radiation (W/m^2)
        FVC=FVC,  # fractional vegetation cover
        VPD_hPa=VPD_hPa,  # vapor pressure deficit (hPa)
        SVP_hPa=SVP_hPa,  # saturation vapor pressure at given air temperature (hPa)
        Ea_hPa=Ea_hPa,  # actual vapor pressure at air temperature (hPa)
        Estar_hPa=Estar_hPa,  # saturation vapor pressure at surface temperature (hPa)
        gamma_hPa=gamma_hPa # psychrometric constant (hPa/°C)
    )

    # calculate soil heat flux
    G = calculate_SEBAL_soil_heat_flux(
        ST_C=ST_C,  # Surface temperature in Celsius
        NDVI=NDVI,  # Normalized Difference Vegetation Index
        albedo=albedo,  # Albedo of the surface
        Rn=Rn_Wm2  # Net radiation (W/m^2)
    )

    # get phi with new comp
    phi = Rn_Wm2 - G

    Es = rt.where((Ep_PT > phi) & (dTS_C > 0) & (Td_C <= 0), Ea_hPa + SMrz * (Estar_hPa - Ea_hPa),
                    Ea_hPa + Ms * (Estar_hPa - Ea_hPa))
    
    # Return all the created variables
    return SM, SMrz, Ms, s1, s3, Ep_PT, Rn_soil, LWnet, G, Tsd_C, Ds, Es, phi



from typing import Union, Tuple
import numpy as np

import rasters as rt
from rasters import Raster

from .constants import *

def initialize_without_solar(
            ST_C: Union[Raster, np.ndarray],  # Surface temperature in Celsius
            Ta_C: Union[Raster, np.ndarray],  # Air temperature in Celsius
            dTS: Union[Raster, np.ndarray],  # Temperature difference between surface and air in Celsius
            Td_C: Union[Raster, np.ndarray],  # Dewpoint temperature in Celsius
            Ea_hPa: Union[Raster, np.ndarray],  # Actual vapor pressure in hPa
            Estar_hPa: Union[Raster, np.ndarray],  # Saturation vapor pressure at surface temperature in hPa
            SVP_hPa: Union[Raster, np.ndarray],  # Saturation vapor pressure at the surface in hPa
            delta_hPa: Union[Raster, np.ndarray],  # Slope of the saturation vapor pressure-temperature curve in hPa
            phi_Wm2: Union[Raster, np.ndarray],  # Available energy in W/m2
            gamma_hPa: Union[Raster, np.ndarray, float] = GAMMA_HPA,  # Psychrometric constant in hPa/°C
            alpha: Union[Raster, np.ndarray, float] = PT_ALPHA  # Priestley-Taylor alpha
      ) -> Tuple[Union[Raster, np.ndarray]]:
      """
      Initializes the variables related to moisture and vapor pressure without considering solar radiation.

      Parameters:
          ST_C (np.ndarray): Surface temperature in Celsius.
          Ta_C (np.ndarray): Air temperature in Celsius.
          dTS (np.ndarray): Temperature difference between surface and air in Celsius.
          Td_C (np.ndarray): Dewpoint temperature in Celsius.
          Ea_hPa (np.ndarray): Actual vapor pressure in hPa.
          Estar (np.ndarray): saturation vapor pressure at surface temperature (hPa/K)
          SVP_hPa (np.ndarray): Saturation vapor pressure at the surface in hPa.
          delta (np.ndarray): Slope of the saturation vapor pressure-temperature curve in hPa/K.
          phi (np.ndarray): available energy in W/m2.

      Returns:
          Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
              - SM (np.ndarray): Soil moisture.
              - SMrz (np.ndarray): Rootzone moisture.
              - s1 (np.ndarray): Slope of saturation vapor pressure and dewpoint temperature.
              - s3 (np.ndarray): Slope of saturation vapor pressure and temperature.
              - s33 (np.ndarray): Slope of saturation vapor pressure and surface temperature.
              - s44 (np.ndarray): s44    
              - Ms (np.ndarray): Surface moisture.
              - Tsd_C (np.ndarray): Surface dewpoint temperature in Celsius.
              - Es (np.ndarray): Surface vapor pressure in hPa.
              - Ds (np.ndarray): Vapor pressure deficit at the surface.
      """
      s33 = (45.03 + 3.014 * ST_C + 0.05345 * ST_C ** 2 + 0.00224 * ST_C ** 3) * 1e-2  # hpa/K
      s1 = (45.03 + 3.014 * Td_C + 0.05345 * Td_C ** 2 + 0.00224 * Td_C ** 3) * 1e-2  # hpa/K
      
      # Surface dewpoint (Celsius)
      Tsd_C = ((Estar_hPa - Ea_hPa) - (s33 * ST_C) + (s1 * Td_C)) / (s1 - s33)
      
      # slope of saturation vapor pressure and temperature
      s3 = rt.where((dTS > -20) & (dTS < 5), (Estar_hPa - Ea_hPa) / (ST_C - Td_C),
            (45.03 + 3.014 * ST_C + 0.05345 * ST_C ** 2 + 0.00224 * ST_C ** 3) * 1e-2)  # hpa/K
      
      # Surface Moisture (Ms)
      # Surface wetness
      Ms = (s1 / s3) * ((Tsd_C - Td_C) / (ST_C - Td_C))
      Ms = rt.clip(rt.where((dTS < 0) & (Ms < 0) & (phi_Wm2 < 0), np.abs(Ms), Ms), 0, 1)

      # Rootzone Moisture (Mrz)
      s44 = (SVP_hPa - Ea_hPa) / (Ta_C - Td_C)
      SMrz = (gamma_hPa * s1 * (Tsd_C - Td_C)) / (
            delta_hPa * s3 * (ST_C - Td_C) + gamma_hPa * s44 * (Ta_C - Td_C) - delta_hPa * s1 * (
            Tsd_C - Td_C))  # rootzone wetness
      SMrz = rt.clip(rt.where((dTS < 0) & (SMrz < 0) & (phi_Wm2 < 0), np.abs(SMrz), SMrz), 0, 1)

      # now the limits of both Ms and Mrz are consistent
      # combine M to account for Hysteresis and initial estimation of surface vapor pressure
      # Potential evaporation (Priestley-Taylor eqn.)
      Ep_PT = (alpha * delta_hPa * phi_Wm2) / (delta_hPa + gamma_hPa)
      Es = rt.where((Ep_PT > phi_Wm2) & (dTS > 0) & (Td_C <= 0), Ea_hPa + SMrz * (Estar_hPa - Ea_hPa),
            Ea_hPa + Ms * (Estar_hPa - Ea_hPa))

      # soil moisture
      SM = rt.where((Ep_PT > phi_Wm2) & (dTS > 0) & (Td_C <= 0), SMrz, Ms)

      # hysteresis logic
      # vapor pressure deficit at surface (Ds is later replaced by D0)
      Ds = (Estar_hPa - Es)

      return SM, SMrz, s1, s3, s33, s44, Ms, Tsd_C, Es, Ds



from typing import Union
import numpy as np

from rasters import Raster

from .constants import SB_SIGMA

def calculate_net_longwave_radiation(
        Ta_C: Union[Raster, np.ndarray], 
        Ea_hPa: Union[Raster, np.ndarray], 
        ST_C: Union[Raster, np.ndarray], 
        emissivity: Union[Raster, np.ndarray], 
        albedo: Union[Raster, np.ndarray],
        sigma: float = SB_SIGMA) -> Union[Raster, np.ndarray]:
    """
    Calculate the net radiation.

    Parameters:
    Ta_C (np.ndarray): Air temperature in Celsius
    Ea_hPa (np.ndarray): Actual vapor pressure at air temperature in hPa
    ST_C (np.ndarray): Surface temperature in Celsius
    emissivity (np.ndarray): Emissivity of the surface
    RG (np.ndarray): Global radiation
    albedo (np.ndarray): Albedo of the surface

    Returns:
    Lnet (np.ndarray): Net longwave radiation
    """
    etaa = 1.24 * (Ea_hPa / (Ta_C + 273.15)) ** (1.0 / 7.0)  # air emissivity
    LWin = sigma * etaa * (Ta_C + 273.15) ** 4
    LWout = sigma * emissivity * (ST_C + 273.15) ** 4
    
    # emissivity was being applied twice here
    # LWnet = emissivity * LWin - LWout
    
    LWnet = LWin - LWout

    return LWnet



from typing import Union, Callable
import logging
from datetime import datetime, timedelta
from os.path import join, abspath, expanduser
from typing import Dict, List
import numpy as np
import warnings
from .diagnostic import diagnostic
import colored_logging as cl
from .meteorology_conversion import calculate_air_density, calculate_specific_heat, calculate_specific_humidity, calculate_surface_pressure, celcius_to_kelvin
import rasters as rt
from geos5fp import GEOS5FP
from .timer import Timer

from rasters import Raster, RasterGeometry

from .vegetation_conversion.vegetation_conversion import FVC_from_NDVI, LAI_from_NDVI

from .constants import *
from .closure import STIC_closure
from .soil_moisture_initialization import initialize_soil_moisture
from .soil_moisture_iteration import iterate_soil_moisture
from .net_radiation import calculate_net_longwave_radiation
from .initialize_with_solar import initialize_with_solar
from .canopy_air_stream import calculate_canopy_air_stream_vapor_pressure
from .initialize_without_solar import initialize_without_solar
from .iterate_with_solar import iterate_with_solar
from .iterate_without_solar import iterate_without_solar
from .root_zone_initialization import calculate_root_zone_moisture

from .soil_heat_flux import calculate_SEBAL_soil_heat_flux

__author__ = 'Kaniska Mallick, Madeleine Pascolini-Campbell, Gregory Halverson'

logger = logging.getLogger(__name__)

LE_CONVERGENCE_TARGET_WM2 = 2.0
MAX_ITERATIONS = 30
USE_VARIABLE_ALPHA = True
SHOW_DISTRIBUTIONS = True

def STIC(
        hour_of_day: Union[Raster, np.ndarray],  # hour of day
        ST_C: Union[Raster, np.ndarray],
        emissivity: Union[Raster, np.ndarray],
        NDVI: Union[Raster, np.ndarray],
        albedo: Union[Raster, np.ndarray],
        Ta_C: Union[Raster, np.ndarray],
        RH: Union[Raster, np.ndarray],
        Rn_Wm2: Union[Raster, np.ndarray],
        geometry: RasterGeometry = None,
        G: Union[Raster, np.ndarray] = None,
        G_method: str = DEFAULT_G_METHOD,
        SM: Union[Raster, np.ndarray] = None,
        Rg_Wm2: Union[Raster, np.ndarray] = None,
        FVC: Union[Raster, np.ndarray] = None,
        LAI: Union[Raster, np.ndarray] = None,
        elevation_m: Union[Raster, np.ndarray] = None,
        delta_hPa: Union[Raster, np.ndarray] = None,
        gamma_hPa: Union[Raster, np.ndarray, float] = GAMMA_HPA,
        rho_kgm3: Union[Raster, np.ndarray] = RHO_KGM3,
        Cp_Jkg: Union[Raster, np.ndarray] = CP_JKG,
        alpha: float = PT_ALPHA,
        LE_convergence_target: float = LE_CONVERGENCE_TARGET_WM2,
        max_iterations: int = MAX_ITERATIONS,
        diagnostic_directory: str = None,
        show_distributions: bool = SHOW_DISTRIBUTIONS,
        use_variable_alpha: bool = USE_VARIABLE_ALPHA) -> Dict[str, Union[Raster, np.ndarray]]:
    results = {}

    if geometry is None and isinstance(ST_C, Raster):
        geometry = ST_C.geometry

    diag_kwargs = {
        "show_distributions": show_distributions, 
        "output_directory": diagnostic_directory
    }

    seconds_of_day = hour_of_day * 3600.0

    # calculate fraction of vegetation cover if it's not given
    if FVC is None:
        FVC = FVC_from_NDVI(NDVI)
    
    # calculate leaf area index if it's not given
    if LAI is None:
        LAI = LAI_from_NDVI(NDVI)

    # saturation air pressure in hPa
    SVP_hPa = 6.13753 * (np.exp((17.27 * Ta_C) / (Ta_C + 237.3)))

    # calculate delta term if it's not given
    if delta_hPa is None:
        # slope of saturation vapor pressure to air temperature (hpa/K)
        delta_hPa = 4098 * SVP_hPa / (Ta_C + 237.3) ** 2

    Ta_K = celcius_to_kelvin(Ta_C)

    # actual vapor pressure at TA (hpa/K)
    Ea_hPa = SVP_hPa * (RH)
    Ea_Pa = Ea_hPa * 100.0

    # vapor pressure deficit (hPa)
    VPD_hPa = SVP_hPa - Ea_hPa

    # swapping in the dew-point calculation from PT-JPL
    Td_C = Ta_C - ((100 - RH * 100) / 5.0)

    # difference between surface and air temperature (Celsius)
    dTS_C = ST_C - Ta_C

    # saturation vapor pressure at surface temperature (hPa/K)
    Estar_hPa = 6.13753 * np.exp((17.27 * ST_C) / (ST_C + 237.3))

    if Rg_Wm2 is None:
        # if G is None and SM is None:
        #     raise ValueError("soil heat flux or soil moisture prior required if solar radiation is not given")

        if G is None:
            G = calculate_SEBAL_soil_heat_flux(
                ST_C=ST_C,
                NDVI=NDVI,
                albedo=albedo,
                Rn=Rn_Wm2,
            )
        
        phi_Wm2 = Rn_Wm2 - G

        # initialize without solar radiation
        SM, SMrz, s1, s3, s33, s44, Ms, Tsd_C, Es_hPa, Ds = initialize_without_solar(
            ST_C = ST_C,  # Surface temperature in Celsius
            Ta_C = Ta_C,  # Air temperature in Celsius
            dTS = dTS_C,  # Temperature difference between surface and air in Celsius
            Td_C = Td_C,  # Dewpoint temperature in Celsius
            Ea_hPa = Ea_hPa,  # Actual vapor pressure in hPa
            Estar_hPa = Estar_hPa,  # Saturation vapor pressure at surface temperature (hPa/K)
            SVP_hPa = SVP_hPa,  # Saturation vapor pressure at the surface in hPa
            delta_hPa = delta_hPa,  # Slope of the saturation vapor pressure-temperature curve in hPa/K
            phi_Wm2 = phi_Wm2,  # Available energy in W/m2
            gamma_hPa = gamma_hPa,  # Psychrometric constant in hPa/°C
            alpha = alpha  # Priestley-Taylor alpha
        )
    else:
        SM, SMrz, Ms, s1, s3, Ep_PT, Rnsoil, LWnet_Wm2, G, Tsd_C, Ds, Es_hPa, phi_Wm2 = initialize_with_solar(
            seconds_of_day = seconds_of_day,  # time of day in seconds since midnight
            Rg_Wm2 = Rg_Wm2,  # solar radiation (W/m^2)
            Rn_Wm2 = Rn_Wm2,  # net radiation (W/m^2)
            ST_C = ST_C,  # surface temperature (Celsius)
            emissivity = emissivity,  # emissivity of the surface
            Ta_C = Ta_C,  # air temperature (Celsius)
            dTS_C = dTS_C,  # surface air temperature difference (Celsius)
            Td_C = Td_C,  # dew point temperature (Celsius)
            VPD_hPa = VPD_hPa,  # vapor pressure deficit (hPa)
            SVP_hPa = SVP_hPa,  # saturation vapor pressure at given air temperature (hPa)
            Ea_hPa = Ea_hPa,  # actual vapor pressure at air temperature (hPa)
            Estar_hPa = Estar_hPa,  # saturation vapor pressure at surface temperature (hPa)
            delta_hPa = delta_hPa,  # slope of saturation vapor pressure to air temperature (hpa/K)
            NDVI=NDVI,  # normalized difference vegetation index
            FVC = FVC,  # fractional vegetation cover
            LAI = LAI,  # leaf area index
            albedo = albedo,  # albedo of the surface
            gamma_hPa=gamma_hPa,  # psychrometric constant (hPa/°C)
            G_method = DEFAULT_G_METHOD,  # method for calculating soil heat flux
        )
    
    diagnostic(Ms, "Ms", **diag_kwargs)

    # STIC analytical equations (convergence on LE)
    gB_ms, gS_ms, dT_C, EF = STIC_closure(
        delta_hPa=delta_hPa, 
        phi_Wm2=phi_Wm2, 
        Es_hPa=Es_hPa, 
        Ea_hPa=Ea_hPa, 
        Estar_hPa=Estar_hPa, 
        SM=SM,
        gamma_hPa=gamma_hPa,
        rho_kgm3=rho_kgm3,
        Cp_Jkg=Cp_Jkg,
        alpha=alpha
    )
    
    gBB = gB_ms
    gSS = gS_ms
    gBB_by_gSS = rt.where(gSS == 0, 0, gBB / gSS)
    gB_by_gS = rt.where(gS_ms == 0, 0, gB_ms / gS_ms)
    dT_C = dT_C
    T0_C = dT_C + Ta_C
    
    PET_Wm2 = ((delta_hPa * phi_Wm2 + rho_kgm3 * Cp_Jkg * gB_ms * VPD_hPa) / (delta_hPa + gamma_hPa))  # Penman potential evaporation
    
    gR = (4 * SB_SIGMA * (Ta_C + 273) ** 3 * emissivity) / (rho_kgm3 * Cp_Jkg)
    omega = ((delta_hPa / gamma_hPa) + 1) / ((delta_hPa / gamma_hPa) + 1 + gB_by_gS)
    LE_eq = (phi_Wm2 * (delta_hPa / gamma_hPa)) / ((delta_hPa / gamma_hPa) + 1)
    LE_imp = (Cp_Jkg * 0.0289644 / gamma_hPa) * gS_ms * 40 * VPD_hPa
    LE_init = omega * LE_eq + (1 - omega) * LE_imp
    dry = (Ds > VPD_hPa) & (PET_Wm2 > phi_Wm2) & (dTS_C > 0) & (Td_C <= 0)
    omega = rt.where(dry,
                        ((delta_hPa / gamma_hPa) + 1 + gR / gB_ms) / ((delta_hPa / gamma_hPa) + 1 + gB_ms / gS_ms + gR / gS_ms + gR / gB_ms),
                        omega)
    LE_eq = rt.where(dry, (phi_Wm2 * (delta_hPa / gamma_hPa)) / ((delta_hPa / gamma_hPa) + 1 + gR / gB_ms), LE_eq)
    LE_init = rt.where(dry, omega * LE_eq + (1 - omega * LE_imp), LE_init)
    
    # sensible heat flux
    H_Wm2 = ((gamma_hPa * phi_Wm2 * (1 + gB_by_gS) - rho_kgm3 * Cp_Jkg * gB_ms * VPD_hPa) / (delta_hPa + gamma_hPa * (1 + (gB_by_gS))))
    
    LE_Wm2_new = LE_init
    LE_Wm2_change = LE_convergence_target
    LE_Wm2_old = LE_Wm2_new
    LE_transpiration_Wm2 = None
    PT_Wm2 = None
    iteration = 1
    LE_Wm2_max_change = 0
    t = Timer()

    while (np.nanmax(LE_Wm2_change) >= LE_convergence_target and iteration <= max_iterations):
        logger.info(f"running STIC iteration {cl.val(iteration)} / {cl.val(max_iterations)}")

        if Rg_Wm2 is None:
            SM, SMrz, Ms, s1, e0, e0star, Tsd_C, D0, alphaN = iterate_without_solar(
                LE = LE_Wm2_new,  # Latent heat flux (W/m^2)
                PET = PET_Wm2,  # Potential evapotranspiration (W/m^2)
                SM = SM,
                ST_C = ST_C,  # Surface temperature (°C)
                Ta_C = Ta_C,  # Air temperature (°C)
                dTS = dTS_C,  # Surface-air temperature difference (°C)
                T0 = T0_C,  # Reference temperature (°C)
                gB = gB_ms,  # Boundary layer conductance (m/s)
                gS = gS_ms,  # Stomatal conductance (m/s)
                Ea_hPa = Ea_hPa,  # Actual vapor pressure (hPa)
                Td_C = Td_C,  # Dew point temperature (°C)
                VPD_hPa = VPD_hPa,  # Vapor pressure deficit (hPa)
                Estar = Estar_hPa,  # Saturation vapor pressure at surface temperature (hPa)
                delta = delta_hPa,  # Slope of the saturation vapor pressure-temperature curve (hPa/°C)
                phi = phi_Wm2,  # available energy (W/m^2)
                Ds = Ds,  # Vapor pressure deficit at source (hPa)
                Es = Es_hPa,  # Saturation vapor pressure (hPa)
                s3 = s3,  # Slope of the saturation vapor pressure and temperature 
                s4 = s44,  # Slope of the saturation vapor pressure and temperature 
                gB_by_gS = gB_by_gS,  # Ratio of boundary layer conductance to stomatal conductance
                gamma_hPa = gamma_hPa,  # Psychrometric constant (hPa/°C)
                rho_kgm3 = rho_kgm3,  # Air density (kg/m^3)
                Cp_Jkg = Cp_Jkg  # Specific heat at constant pressure (J/kg/K)
            )
        else:
            SM, G, e0, e0star, D0, alphaN = iterate_with_solar(
                seconds_of_day = seconds_of_day,  # Seconds of the day
                ST_C = ST_C,  # Soil temperature (°C)
                NDVI = NDVI,  # Normalized Difference Vegetation Index
                albedo = albedo,  # Albedo
                gB_ms = gB_ms,  # boundary layer conductance (m/s)
                gS_ms = gS_ms,  # stomatal conductance (m/s)
                LE_Wm2 = LE_Wm2_new,  # latent heat flux (W/m^2)
                Rg_Wm2 = Rg_Wm2,  # Incoming solar radiation (W/m^2)
                Rn_Wm2 = Rn_Wm2,  # Net radiation (W/m^2)
                LWnet_Wm2 = LWnet_Wm2,  # Net longwave radiation (W/m^2)
                Ta_C = Ta_C,  # Air temperature (°C)
                dTS_C = dTS_C,  # Change in soil temperature (°C)
                Td_C = Td_C,  # Dew point temperature (°C)
                Tsd_C = Tsd_C,  # Soil dew point temperature (°C)
                Ea_hPa = Ea_hPa,  # actual vapor pressure (hPa)
                Estar_hPa = Estar_hPa,  # saturation vapor pressure at surface temperature (hPa)
                VPD_hPa = VPD_hPa,  # Vapor pressure deficit (hPa)
                SVP_hPa = SVP_hPa,  # Saturation vapor pressure (hPa)
                delta_hPa = delta_hPa,  # Slope of the saturation vapor pressure-temperature curve (hPa/°C)
                phi_Wm2 = phi_Wm2,  # Net radiation minus soil heat flux (W/m^2)
                Es_hPa = Es_hPa,  # Saturation vapor pressure (hPa)
                s1 = s1,  # Soil moisture parameter
                s3 = s3,  # Soil moisture parameter
                FVC = FVC,  # Fractional canopy cover
                T0_C = T0_C,  # Reference temperature (°C)
                gB_by_gS = gB_by_gS,  # Ratio of boundary layer conductance to stomatal conductance
                gamma_hPa = gamma_hPa,  # Psychrometric constant (hPa/°C)
                rho_kgm3 = rho_kgm3,  # Air density (kg/m^3)
                Cp_Jkg = Cp_Jkg,  # Specific heat at constant pressure (J/kg/K)
                G_method = "santanello"  # Method for calculating soil heat flux
            )

        if use_variable_alpha:
            alpha = alphaN
            logger.info(f"using variable Priestley-Taylor alpha with mean: {cl.val(np.round(np.nanmean(alpha), 3))}")

        # re-estimated conductances and states
        gB_ms, gS_ms, dT_C, EF = STIC_closure(
            delta_hPa=delta_hPa,  # Slope of the saturation vapor pressure-temperature curve (hPa/°C)
            phi_Wm2=phi_Wm2,  # available energy (W/m^2)
            Es_hPa=Es_hPa,  # Vapor pressure at the reference height (hPa)
            Ea_hPa=Ea_hPa,  # Actual vapor pressure (hPa)
            Estar_hPa=Estar_hPa,  # Saturation vapor pressure at the reference height (hPa)
            SM=SM,  # Soil moisture (m³/m³)
            gamma_hPa=gamma_hPa,  # Psychrometric constant (hPa/°C)
            rho_kgm3=rho_kgm3,  # Air density (kg/m³)
            Cp_Jkg=Cp_Jkg,  # Specific heat capacity of air (J/kg/°C)
            alpha=alpha  # Stability correction factor for conductance 
        )

        gB_by_gS = rt.where(gS_ms == 0, 0, gB_ms / gS_ms)
        T0_C = dT_C + Ta_C
        # latent heat flux
        LE_Wm2_new = ((delta_hPa * phi_Wm2 + rho_kgm3 * Cp_Jkg * gB_ms * VPD_hPa) / (delta_hPa + gamma_hPa * (1 + gB_by_gS)))
        LE_Wm2_new = rt.where(LE_Wm2_new > phi_Wm2, phi_Wm2, LE_Wm2_new)
        # Sensible Heat Flux
        H_Wm2 = ((gamma_hPa * phi_Wm2 * (1 + gB_by_gS) - rho_kgm3 * Cp_Jkg * gB_ms * VPD_hPa) / (delta_hPa + gamma_hPa * (1 + (gB_by_gS))))
        # potential evaporation (Penman)
        PET_Wm2 = ((delta_hPa * phi_Wm2 + rho_kgm3 * Cp_Jkg * gB_ms * VPD_hPa) / (delta_hPa + gamma_hPa))
        # Potential Transpiration
        PT_Wm2 = (delta_hPa * phi_Wm2 + rho_kgm3 * Cp_Jkg * gB_ms * VPD_hPa) / (delta_hPa + gamma_hPa * (1 + SM * gB_by_gS))  # potential transpiration
        # ET PARTITIONING
        LE_soil_Wm2 = rt.clip(SM * PET_Wm2, 0, None)
        LE_transpiration_Wm2 = rt.clip(LE_Wm2_new - LE_soil_Wm2, 0, None)
        # change in latent heat flux estimate
        LE_Wm2_change = np.abs(LE_Wm2_old - LE_Wm2_new)
        LE_Wm2_new = rt.where(np.isnan(LE_Wm2_new), LE_Wm2_old, LE_Wm2_new)
        LE_Wm2_old = LE_Wm2_new
        LE_Wm2_max_change = np.nanmax(LE_Wm2_change)
        logger.info(
            f"completed STIC iteration {cl.val(iteration)} / {cl.val(max_iterations)} with max LE change: {cl.val(np.round(LE_Wm2_max_change, 3))} ({t} seconds)")
        
        diagnostic(SM, f"SM_{iteration}", **diag_kwargs)
        diagnostic(G, f"G_{iteration}", **diag_kwargs)
        diagnostic(LE_Wm2_new, f"LE_{iteration}", **diag_kwargs)

        if LE_Wm2_max_change <= LE_convergence_target:
            logger.info(f"max LE change {cl.val(np.round(LE_Wm2_max_change, 3))} within convergence target {cl.val(np.round(LE_convergence_target, 3))} with {cl.val(iteration)} iteration{'s' if iteration > 1 else ''}")

        iteration += 1

    iteration -= 1
    results["LE_max_change"] = LE_Wm2_max_change
    results["iteration"] = iteration

    LE = LE_Wm2_new

    results["LE"] = LE
    results["LE_change"] = LE_Wm2_change
    results["LEt"] = LE_transpiration_Wm2
    results["PT"] = PT_Wm2
    results["PET"] = PET_Wm2
    results["G"] = G

    if geometry is not None:
        for name, array in results.items():
            try:
                results[name] = Raster(array.reshape(geometry.shape), geometry=geometry)
            except Exception as e:
                pass

    warnings.resetwarnings()

    return results

# def process_STIC_raster(
#         geometry: RasterGeometry,
#         hour_of_day: float,
#         ST_C: Raster,
#         emissivity: Raster,
#         NDVI: Raster,
#         albedo: Raster,
#         Ta_C: Raster,
#         RH: Raster,
#         Rn: Raster,
#         G: Raster = None,
#         G_method: str = DEFAULT_G_METHOD,
#         SM: Raster = None,
#         Rg: Raster = None,
#         LE_convergence_target: float = LE_CONVERGENCE_TARGET_WM2,
#         diagnostic_directory: str = None,
#         max_iterations: int = MAX_ITERATIONS) -> Dict[str, Raster]:
#     results = process_STIC_array(
#         hour_of_day=hour_of_day,
#         ST_C=ST_C,
#         emissivity=emissivity,
#         NDVI=NDVI,
#         albedo=albedo,
#         Ta_C=Ta_C,
#         RH=RH,
#         Rn_Wm2=Rn,
#         G=G,
#         G_method=G_method,
#         SM=SM,
#         Rg_Wm2=Rg,
#         LE_convergence_target=LE_convergence_target,
#         diagnostic_directory=diagnostic_directory,
#         max_iterations=max_iterations
#     )
    
#     for name, array in results.items():
#         try:
#             results[name] = Raster(array.reshape(geometry.shape), geometry=geometry)
#         except Exception as e:
#             pass
    
#     return results



RHO_KGM3 = 1.2  # Air density (kg m-3)
CP_JKG = 1013  # Specific heat of air at constant pressure (J/kg/K)
GAMMA_HPA = 0.67  # Psychrometric constant (hpa/K)
PT_ALPHA = 1.26  # Priestley-Taylor coefficient
SB_SIGMA = 5.67e-8  # Stefann Boltzmann constant
DEFAULT_G_METHOD = "santanello"
